import numpy as np

from .fourier_transform import FourierTransform, multiplex_for_tensor_fields, _get_float_and_complex_dtype
from ..field import Field
from ..config import Configuration

class NaiveFourierTransform(FourierTransform):
	'''The naive Fourier transform (NFT).

	This Fourier transform can operate on any types of coordinates, and calculates
	the Fourier integral using brute force. This class is written as generally as
	possible to act as a "ground truth" for the other Fourier transforms. As such,
	it has terrible performance.

	.. note::
		In almost all cases, you should not use this class. The FastFourierTransform and
		MatrixFourierTransform classes, if applicable, are multiple orders of magnitudes faster
		and use multiple orders of magnitude less memory. This class should only be used in a last
		resort and if you know what you are doing.

	.. note::
		The transformation matrices can be very large, quickly overwhelming any computer even
		for relatively small input or output grids.

	Parameters
	----------
	input_grid : Grid
		The grid that is expected for the input field.
	output_grid : Grid
		The grid that is produced by the Fourier transform.
	precompute_matrices : boolean or None
		Whether to precompute the matrices used in the NFT. Turning this on will cache calculated matrices
		for future evaluations of the Fourier transform. This will use a large amount of memory. If this is
		None, the choice will be determined by the configuration file.

	Raises
	------
	ValueError
		If the output grid has a different dimension than the input grid.
	'''
	def __init__(self, input_grid, output_grid, precompute_matrices=None):
		if input_grid.ndim != output_grid.ndim:
			raise ValueError('The input_grid must have the same dimensions as the output_grid.')

		self.input_grid = input_grid
		self.output_grid = output_grid

		self._T_forward = None
		self._T_backward = None

		if precompute_matrices is None:
			precompute_matrices = Configuration().fourier.nft.precompute_matrices
		self.precompute_matrices = precompute_matrices

		self.coords_in = np.array(self.input_grid.as_('cartesian').coords)
		self.coords_out = np.array(self.output_grid.as_('cartesian').coords)

	@property
	def T_forward(self):
		'''The cached forward propagation matrix.
		'''
		if not self.precompute_matrices:
			return self.get_transformation_matrix_forward()

		if self._T_forward is None:
			self._T_forward = self.get_transformation_matrix_forward()

		return self._T_forward

	@property
	def T_backward(self):
		'''The cached backward propagation matrix.
		'''
		if not self.precompute_matrices:
			return self.get_transformation_matrix_backward()

		if self._T_backward is None:
			self._T_backward = self.get_transformation_matrix_backward()

		return self._T_backward

	@multiplex_for_tensor_fields
	def forward(self, field):
		'''Returns the forward Fourier transform of the :class:`Field` field.

		Parameters
		----------
		field : Field
			The field to Fourier transform.

		Returns
		--------
		Field
			The Fourier transform of the field.
		'''
		if self.precompute_matrices:
			res = self.T_forward.dot(field.ravel())
		else:
			res = np.array([(field * self.input_grid.weights).dot(np.exp(-1j * np.dot(p, self.coords_in))) for p in self.coords_out.T])

		float_dtype, complex_dtype = _get_float_and_complex_dtype(field.dtype)
		return Field(res, self.output_grid).astype(complex_dtype, copy=False)

	@multiplex_for_tensor_fields
	def backward(self, field):
		'''Returns the inverse Fourier transform of the :class:`Field` field.

		Parameters
		----------
		field : Field
			The field to inverse Fourier transform.

		Returns
		--------
		Field
			The inverse Fourier transform of the field.
		'''
		if self.precompute_matrices:
			res = self.T_backward.dot(field.ravel())
		else:
			res = np.array([(field * self.output_grid.weights).dot(np.exp(1j * np.dot(p, self.coords_out))) for p in self.coords_in.T])
			res /= (2 * np.pi)**self.input_grid.ndim

		float_dtype, complex_dtype = _get_float_and_complex_dtype(field.dtype)
		return Field(res, self.input_grid).astype(complex_dtype, copy=False)



import os
import shutil
import time

from nbclient import NotebookClient
from nbclient.exceptions import CellExecutionError
from nbconvert.exporters import RSTExporter
from nbconvert.writers import FilesWriter
import nbformat

from PIL import Image, ImageChops

def compile_tutorial(tutorial_name, force_recompile=False):
	print('- Tutorial "' + tutorial_name + '"')

	notebook_path = 'tutorial_notebooks/' + tutorial_name + '/' + tutorial_name + '.ipynb'
	export_path = 'tutorials/' + tutorial_name + '/' + tutorial_name
	thumb_dest = os.path.dirname(export_path) + '/thumb.png'

	if not os.path.exists(os.path.dirname(export_path)):
		os.makedirs(os.path.dirname(export_path))

	# Read in notebook
	print('  Reading notebook...')
	notebook = nbformat.read(notebook_path, 4)

	# Scrape title, description and thumbnail
	first_cell = notebook.cells[0]

	title = first_cell.source.splitlines()[0]
	if '#' in title:
		title = title.replace('#', '').strip()

	description = ''
	for line in first_cell.source.splitlines()[1:]:
		if line.strip():
			description = line.strip()
			break

	if not description:
		print('  Description could not be found in the notebook.')

	if 'thumbnail_figure_index' in notebook.metadata:
		thumbnail_figure_index = notebook.metadata['thumbnail_figure_index']
	else:
		thumbnail_figure_index = -1

	if 'level' in notebook.metadata:
		level = notebook.metadata['level'].capitalize()
	elif 'difficulty' in notebook.metadata:
		level = notebook.metadata['difficulty'].capitalize()
	else:
		level = 'Unknown'

	# Check if the tutorial was already executed.
	if os.path.exists(export_path + '.rst'):
		if os.path.getmtime(export_path + '.rst') > os.path.getmtime(notebook_path):
			if force_recompile:
				print('  Already compiled. Recompiling anyway...')
			else:
				print('  Already compiled. Skipping compilation...')
				return title, level, description, thumb_dest.split('/', 1)[-1]

	# Execute notebook if not already executed
	already_executed = any(c.get('outputs') or c.get('execution_count') for c in notebook.cells if c.cell_type == 'code')

	resources = {'metadata': {'path': os.path.dirname(notebook_path)}}

	if not already_executed:
		print('  Executing', end='')
		start = time.time()

		additional_cell_1 = {
			"cell_type": "code",
			"execution_count": None,
			"metadata": {},
			"outputs": [],
			"source": r"%matplotlib inline" + '\n' +
				r"%config InlineBackend.print_figure_kwargs = {'bbox_inches': None}"
			}

		additional_cell_2 = {
			"cell_type": "code",
			"execution_count": None,
			"metadata": {},
			"outputs": [],
			"source": "import matplotlib as mpl\nmpl.rcParams['figure.figsize'] = (8, 6)\nmpl.rcParams['figure.dpi'] = 150\nmpl.rcParams['savefig.dpi'] = 150"
			}

		notebook.cells.insert(1, nbformat.from_dict(additional_cell_1))
		notebook.cells.insert(2, nbformat.from_dict(additional_cell_2))

		client = NotebookClient(nb=notebook, resources=resources, timeout=585, kernel_name='python3')

		try:
			with client.setup_kernel():
				for i, cell in enumerate(notebook.cells):
					print('.', end='')

					client.execute_cell(cell, i)

			client.set_widgets_metadata()
		except CellExecutionError as err:
			print('  Error while processing notebook:')
			print('  ', err)

		print('')

		notebook.cells.pop(2)
		notebook.cells.pop(1)

		end = time.time()

		time_taken = end - start
		if time_taken > 60:
			print('  Execution took %dm%02ds.' % (time_taken / 60, time_taken % 60))
		else:
			print('  Execution took %ds.' % time_taken)
	else:
		print('  Notebook was already executed.')

	print('  Rendering tutorial...')
	exporter = RSTExporter()
	output, resources = exporter.from_notebook_node(notebook, resources)

	writer = FilesWriter(build_directory=os.path.dirname(export_path))
	writer.write(output, resources, notebook_name=os.path.basename(export_path))

	pictures = sorted(resources['outputs'], key=output.find)

	try:
		thumbnail_source = pictures[thumbnail_figure_index]

		# Read in thumbnail source image
		img = Image.open(os.path.dirname(export_path) + '/' + thumbnail_source)

		# Trim whitespace
		bg = Image.new(img.mode, img.size, img.getpixel((0, 0)))
		diff = ImageChops.difference(img, bg)
		diff = ImageChops.add(diff, diff)
		bbox = diff.getbbox()
		if bbox:
			img = img.crop(bbox)

		# Resize image to have a width of 400px
		img.thumbnail([400, 1000])

		# Save thumbnail
		img.save(thumb_dest)
	except:
		shutil.copyfile('_static/no_thumb.png', thumb_dest)

	print('  Done!')

	return title, level, description, thumb_dest.split('/', 1)[-1]

index_preamble = '''
Tutorials
=========

These tutorials demonstrate the features of HCIPy in the context of a standard workflow. Tutorials are separated in three categories, depending on the required level of familiarity with HCIPy.
'''

entry_template = '''
.. only:: html

	.. container:: tutorial_item

		:doc:`{title} <{name}/{name}>`

		.. container:: tutorial_row

			.. container:: tutorial_thumbnail

				.. figure:: {thumbnail_file}

			.. container:: tutorial_description

				**Level:** {level}

				**Description:** {description}
'''

beginner_preamble = '''
Beginner
--------

These tutorials provide an introduction to the basic parts of HCIPy. New users should read these tutorials to get started with HCIPy.
'''

intermediate_preamble = '''
Intermediate
------------

These tutorials show the main functionality using the built-in classes of HCIPy. These tutorials focus on one aspect of high-contrast imaging.
'''

advanced_preamble = '''
Advanced
--------
These tutorials show how to use HCIPy for your own research. This includes extending HCIPy with your own optical elements and advanced use cases.
'''

expert_preamble = '''
Expert
------

These tutorials provide examples from actual published research that made heavy use of HCIPy for their optical propagations.
'''

unknown_preamble = '''
Unknown
-------

These tutorials do not have their level of difficulty rated.
'''

level_preambles = [beginner_preamble, intermediate_preamble, advanced_preamble, expert_preamble, unknown_preamble]

def compile_all_tutorials():
	print('Compiling all tutorials...')

	tutorials = {}
	tutorial_names = sorted(os.listdir('tutorial_notebooks/'))
	tutorial_names = [name for name in tutorial_names if 'checkpoint' not in name]

	for name in tutorial_names:
		tutorials[name] = compile_tutorial(name)

	# Sort by level
	levels = ['Beginner', 'Intermediate', 'Advanced', 'Expert', 'Unknown']

	tutorial_names_by_level = [[] for i in range(len(levels))]
	for name in tutorial_names:
		tutorial_names_by_level[levels.index(tutorials[name][1])].append(name)

	f = open('tutorials/index.rst', 'w')
	f.write(index_preamble)

	for preamble, names in zip(level_preambles, tutorial_names_by_level):
		# Don't write this level if there are no tutorials in it.
		if len(names) == 0:
			continue

		f.write(preamble)

		# Write toctree
		f.write('\n.. toctree::\n    :maxdepth: 1\n    :hidden:\n\n')
		for name in names:
			f.write('    ' + name + '/' + name + '\n')
		f.write('\n\n')

		# Write list
		for name in names:
			title, level, desc, thumb = tutorials[name]
			f.write(entry_template.format(thumbnail_file=thumb, title=title, level=level, description=desc, name=name))

	f.close()

if __name__ == '__main__':
	# Compile all tutorials
	import os
	os.chdir(os.path.dirname(os.path.abspath(__file__)))

	import sys
	sys.path.insert(0, '.')

	compile_all_tutorials()



from setuptools import setup, find_packages

# read the contents of the README.md file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md')) as f:
	long_description = f.read()

setup(name='hcipy',
	use_scm_version=True,
	description='A framework for performing optical propagation simulations, meant for high contrast imaging, in Python.',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://github.com/ehpor/hcipy/',
	author='Emiel Por',
	author_email='por@strw.leidenuniv.nl',
	packages=find_packages(),
	package_data={'hcipy': ['data/*']},
	setup_requires=[
		'setuptools_scm'],
	install_requires=[
		"numpy",
		"scipy",
		"matplotlib>=2.0.0",
		"Pillow",
		"pyyaml",
		"astropy",
		"imageio",
		"xxhash",
		"numexpr",
		"asdf"],
	extras_require={
		"dev": [
			"pytest",
			"codecov",
			"coverage",
			"mpmath",
			"dill"],
		"doc": [
			"numpydoc",
			"sphinx_rtd_theme",
			"nbsphinx",
			"jupyter_client",
			"ipykernel",
			"poppy",
			"nbclient",
			"nbformat",
			"nbconvert",
			"sphinx-automodapi",
			"progressbar2"]},
	zip_safe=False,
	classifiers=[
		"Development Status :: 3 - Alpha",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.5",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Topic :: Scientific/Engineering :: Astronomy"],
	license_file="LICENSE"
)



import numpy as np

from .grid import Grid
from .coordinates import UnstructuredCoords

class PolarGrid(Grid):
	'''A grid representing a two-dimensional Polar coordinate system.
	'''
	_coordinate_system = 'polar'

	@property
	def r(self):
		'''The radial coordinate (dimension 0).
		'''
		return self.coords[0]

	@property
	def theta(self):
		''' The angular coordinate (dimension 1).
		'''
		return self.coords[1]

	def scale(self, scale):
		'''Scale the grid in-place.

		Parameters
		----------
		scale : array_like
			The factor with which to scale the grid.

		Returns
		-------
		Grid
			Itself to allow for chaining these transformations.
		'''
		self.coords *= np.array([scale, 1])
		self.weights *= np.abs(scale)**(self.ndim)
		return self

	def shift(self, shift):
		'''Shift the grid in-place.

		.. caution::
		   All structure in the coordinates will be destroyed.

		Parameters
		----------
		shift : array_like
			The amount with which to shift the grid.

		Returns
		-------
		Grid
			Itself to allow for chaining these transformations.
		'''
		new_grid = PolarGrid(UnstructuredCoords(self.coords)).as_('cartesian')
		self.coords = new_grid.shifted(shift).as_('polar').coords

		return self

	def shifted(self, shift):
		'''A shifted copy of this grid.

		.. caution::
		   The returned grid is a Cartesian grid.

		Parameters
		----------
		shift : array_like
			The amount with which to shift the grid.

		Returns
		-------
		Grid
			The scaled grid.
		'''
		grid = self.as_('cartesian')
		grid.shift(shift)
		return grid

	def rotate(self, angle, axis=None):
		'''Rotate the grid in-place.

		Parameters
		----------
		angle : scalar
			The angle in radians.
		axis : ndarray or None
			The axis of rotation. For this (polar) grid, it is ignored.

		Returns
		-------
		Grid
			Itself to allow for chaining these transformations.
		'''
		self.theta += angle
		return self

	@staticmethod
	def _get_automatic_weights(coords):
		return None

def _cartesian_to_polar(self):
	from .coordinates import UnstructuredCoords

	x = self.x
	y = self.y
	r = np.hypot(x, y)
	theta = np.arctan2(y, x)
	return PolarGrid(UnstructuredCoords([r, theta]))

def _polar_to_cartesian(self):
	from .coordinates import UnstructuredCoords
	from .cartesian_grid import CartesianGrid

	r = self.r
	theta = self.theta
	x = r * np.cos(theta)
	y = r * np.sin(theta)
	return CartesianGrid(UnstructuredCoords([x, y]))

Grid._add_coordinate_system('polar', PolarGrid)

Grid._add_coordinate_system_transformation('cartesian', 'polar', _cartesian_to_polar)
Grid._add_coordinate_system_transformation('polar', 'cartesian', _polar_to_cartesian)



from scipy.interpolate import RegularGridInterpolator, LinearNDInterpolator
import numpy as np
from ..field import Field

def make_linear_interpolator_separated(field, grid=None, fill_value=np.nan):
	'''Make a linear interpolators for a separated grid.

	Parameters
	----------
	field : Field or ndarray
		The field to interpolate.
	grid : Grid or None
		The grid of the field. If it is given, the grid of `field` is replaced by this grid.
	fill_value : scalar
		The value to use for points outside of the domain of the input field. If this is None, the values
		outside the domain are extrapolated.

	Returns
	-------
	Field generator
		The interpolator, as a Field generator. The grid on which this field generator will evaluated, does
		not have to be separated.
	'''
	if grid is None:
		grid = field.grid
	else:
		field = Field(field, grid)

	axes_reversed = np.array(grid.separated_coords)
	interp = RegularGridInterpolator(axes_reversed, field.shaped, 'linear', False, fill_value)

	def interpolator(evaluated_grid):
		evaluated_coords = np.flip(np.array(evaluated_grid.coords), 0)
		res = interp(evaluated_coords.T)
		return Field(res.ravel(), evaluated_grid)

	return interpolator

def make_linear_interpolator_unstructured(field, grid=None, fill_value=np.nan):
	'''Make a linear interpolator for an unstructured grid.

	Parameters
	----------
	field : Field or array_like
		The field to interpolate.
	grid : Grid or None
		The grid of the field. If it is given, the grid of `field` is replaced by this grid.
	fill_value : scalar
		The value to use for points outside of the domain of the input field. Extrapolation is not supported.

	Returns
	-------
	Field generator
		The interpolator as a Field generator. The grid on which this field generator will be evaluated does
		not need to have any structure.
	'''
	if fill_value is None:
		raise ValueError('Extrapolation is not supported for a linear interpolator on an unstructured grid.')

	if grid is None:
		grid = field.grid
	else:
		field = Field(field, grid)

	interp = LinearNDInterpolator(grid.points, field, fill_value)

	def interpolator(evaluated_grid):
		res = interp(grid.points)
		return Field(res, evaluated_grid)

	return interpolator

def make_linear_interpolator(field, grid=None, fill_value=None):
	'''Make a linear interpolator for any type of grid.

	Parameters
	----------
	field : Field or array_like
		The field to interpolate.
	grid : Grid or None
		The grid of the field. If it is given, the grid of `field` is replaced by this grid.
	fill_value : scalar or None
		The value to use for points outside of the domain of the input field. Extrapolation is not supported.
		If it is None, a numpy.nan value will be used for points outside of the domain.

	Returns
	-------
	Field generator
		The interpolator as a Field generator. The grid on which this field generator will be evaluated does
		not need to have any structure.
	'''
	if grid is None:
		grid = field.grid

	if grid.is_unstructured:
		return make_linear_interpolator_unstructured(field, grid)
	else:
		return make_linear_interpolator_separated(field, grid)



from __future__ import division

from ..optics import OpticalElement
from ..field import Field
from ..propagation import FresnelPropagator

import numpy as np
from scipy.special import gamma, kv

class AtmosphericLayer(OpticalElement):
	def __init__(self, input_grid, Cn_squared=None, L0=np.inf, velocity=0, height=0):
		'''A single infinitely-thin atmospheric layer.

		This class serves as a base class for all atmospheric layers. Multiply
		atmospheric layers can be combined into an :class:`MultiLayerAtmosphere` which
		provides modelling of scintillation between layers.

		Parameters
		----------
		input_grid : Grid
			The grid on which the wavefront will be defined.
		Cn_squared : scalar
			The integrated value of Cn^2 for the layer.
		L0 : scalar
			The outer scale of the layer.
		velocity : scalar or array_like
			The velocity of the layer. If a scalar is given, its direction is
			chosen randomly.
		height : scalar
			The height of the atmospheric layer above the ground.

		Attributes
		----------
		input_grid : Grid
			The grid on which the wavefront will be defined.
		t : scalar
			The current time.
		Cn_squared : scalar
			The integrated value of Cn^2 for the layer.
		L0 : scalar
			The outer scale of the phase structure function.
		velocity : array_like
			The two-dimensional velocity of the layer.
		height : scalar
			The height of the atmospheric layer above the ground.
		'''
		self.input_grid = input_grid
		self.Cn_squared = Cn_squared
		self.L0 = L0

		self._velocity = None
		self.velocity = velocity

		self.height = height
		self._t = 0

	def evolve_until(self, t):
		'''Evolve the atmospheric layer until time `t`.

		Parameters
		----------
		t : scalar
			The time to which to evolve the atmospheric layer.
		'''
		raise NotImplementedError()

	def reset(self):
		'''Reset the phase screen.

		This will create a randomized uncorrelated phase screen.
		'''
		raise NotImplementedError()

	@property
	def t(self):
		'''The current time of the atmospheric layer.
		'''
		return self._t

	@t.setter
	def t(self, t):
		self.evolve_until(t)

	@property
	def Cn_squared(self):
		'''The integrated value of Cn^2 for the layer.
		'''
		return self._Cn_squared

	@Cn_squared.setter
	def Cn_squared(self, Cn_squared):
		raise NotImplementedError()

	@property
	def outer_scale(self):
		'''The outer scale of the phase structure function.
		'''
		return self._outer_scale

	@outer_scale.setter
	def outer_scale(self, L0):
		raise NotImplementedError()

	@property
	def L0(self):
		'''The outer scale of the phase structure function.
		'''
		return self.outer_scale

	@L0.setter
	def L0(self, L0):
		self.outer_scale = L0

	@property
	def velocity(self):
		'''The two-dimensional velocity of the layer.
		'''
		return self._velocity

	@velocity.setter
	def velocity(self, velocity):
		if np.isscalar(velocity):
			if self._velocity is not None:
				vel = np.sqrt(np.dot(self._velocity, self._velocity))
				if vel > 0:
					self._velocity *= velocity / vel
					return

			theta = np.random.rand() * 2 * np.pi
			self._velocity = velocity * np.array([np.cos(theta), np.sin(theta)])
		else:
			self._velocity = np.array(velocity)

	def phase_for(self, wavelength):
		'''Get the phase screen in radians at a certain wavelength.

		Each atmospheric layer is modelled as an infinitely-thin phase screen.

		Parameters
		----------
		wavelength : scalar
			The wavelength at which to calculate the phase screen.
		'''
		raise NotImplementedError()

	@property
	def output_grid(self):
		return self.input_grid

	def forward(self, wf):
		wf = wf.copy()
		wf.electric_field *= np.exp(1j * self.phase_for(wf.wavelength))
		return wf

	def backward(self, wf):
		wf = wf.copy()
		wf.electric_field *= np.exp(-1j * self.phase_for(wf.wavelength))
		return wf

class MultiLayerAtmosphere(OpticalElement):
	def __init__(self, layers, scintillation=False, scintilation=None):
		'''A multi-layer atmospheric model.

		This :class:`OpticalElement` can model turbulence and scintillation effects
		due to atmospheric turbulence by propagating light through a series of
		infinitely-thin atmospheric phase screens at different altitudes. The distance
		between two phase screens can be propagated using Fresnel propagation, or using
		no :class:`Propagator`.

		Parameters
		----------
		layers : list of AtmosphericLayer objects
			The series of atmospheric layers in this model.
		scintillation : bool
			If True, then the distance between two phase screens is propagated using
			a :class:`FresnelPropagator`. Otherwise, no propagator will be used.
		'''
		# Retain backwards compatibility.
		if scintilation is not None:
			import warnings
			warnings.warn('Please use the correct spelling for scintillation.')

			scintillation = scintilation

		self.layers = layers
		self._scintillation = scintillation
		self._t = 0
		self._dirty = True

		self.calculate_propagators()

	def calculate_propagators(self):
		'''Recalculates the list of optical elements used for a propagation.

		This function is called automatically by other functions, but a recalculation
		can be forced by calling it explicitly.
		'''
		heights = np.array([l.height for l in self.layers])
		layer_indices = np.argsort(-heights)

		sorted_heights = heights[layer_indices]
		delta_heights = sorted_heights[:-1] - sorted_heights[1:]
		grid = self.layers[0].input_grid

		if self.scintillation:
			propagators = [FresnelPropagator(grid, h) for h in delta_heights]

		self.elements = []
		for i, j in enumerate(layer_indices):
			self.elements.append(self.layers[j])
			if self.scintillation and i < len(propagators):
				self.elements.append(propagators[i])

		if self.scintillation and sorted_heights[-1] > 0:
			self.elements.append(FresnelPropagator(grid, sorted_heights[-1]))

		self._dirty = False

	def reset(self):
		for l in self.layers:
			l.reset()

	@property
	def layers(self):
		'''A list of :class:`AtmosphericLayer` objects.
		'''
		return self._layers

	@layers.setter
	def layers(self, layers):
		self._layers = layers
		self._dirty = True

	def phase_for(self, wavelength):
		'''Get the unwrapped phase in radians for the atmosphere.

		Parameters
		----------
		wavelength : scalar
			The wavelength at which to calculate the phase screen.

		Returns
		-------
		Field
			The total unwrapped phase screen.
		'''
		if self.scintillation:
			raise ValueError('Cannot get the unwrapped phase for an atmosphere with scintillation.')

		unwrapped_phases = [layer.phase_for(wavelength) for layer in self.layers]

		return Field(np.sum(unwrapped_phases, axis=0), unwrapped_phases[-1].grid)

	@property
	def scintillation(self):
		'''Whether to include scintillation effects in the propagation.
		'''
		return self._scintillation

	@scintillation.setter
	def scintillation(self, scintillation):
		self._dirty = scintillation != self.scintillation
		self._scintillation = scintillation

	def evolve_until(self, t):
		'''Evolve all atmospheric layers to a time t.

		Parameters
		----------
		t : scalar
			The time to which to evolve the atmospheric layers.
		'''
		for l in self.layers:
			l.evolve_until(t)
		self._t = t

	@property
	def Cn_squared(self):
		'''The total Cn^2 value of the simulated atmosphere.
		'''
		return np.sum([l.Cn_squared for l in self.layers])

	@Cn_squared.setter
	def Cn_squared(self, Cn_squared):
		old_Cn_squared = self.Cn_squared
		for l in self.layers:
			l.Cn_squared = l.Cn_squared / old_Cn_squared * Cn_squared

	@property
	def outer_scale(self):
		'''The outer scale of all layers.
		'''
		return self.layers[0].outer_scale

	@outer_scale.setter
	def outer_scale(self, L0):
		for l in self.layers:
			l.outer_scale = L0

	@property
	def t(self):
		'''The current time.
		'''
		return self._t

	@t.setter
	def t(self, t):
		self.evolve_until(t)

	def forward(self, wavefront):
		if self._dirty:
			self.calculate_propagators()

		wf = wavefront.copy()
		for el in self.elements:
			wf = el.forward(wf)
		return wf

	def backward(self, wavefront):
		if self._dirty:
			self.calculate_propagators()

		wf = wavefront.copy()
		for el in reversed(self.elements):
			wf = el.backward(wf)
		return wf

def phase_covariance_von_karman(r0, L0):
	'''Return a Field generator for the phase covariance function for Von Karman turbulence.

	Parameters
	----------
	r0 : scalar
		The Fried parameter in meters.
	L0 : scalar
		The outer scale in meters.

	Returns
	-------
	Field generator
		The phase covariance Field generator.
	'''
	def func(grid):
		r = grid.as_('polar').r + 1e-10

		a = (L0 / r0)**(5 / 3)
		b = gamma(11 / 6) / (2**(5 / 6) * np.pi**(8 / 3))
		c = (24 / 5 * gamma(6 / 5))**(5 / 6)
		d = (2 * np.pi * r / L0)**(5 / 6)
		e = kv(5 / 6, 2 * np.pi * r / L0)

		return Field(a * b * c * d * e, grid)
	return func

def phase_structure_function_von_karman(r0, L0):
	'''Return a Field generator for the phase structure function for Von Karman turbulence.

	Parameters
	----------
	r0 : scalar
		The Fried parameter in meters.
	L0 : scalar
		The outer scale in meters.

	Returns
	-------
	Field generator
		The phase structure Field generator.
	'''
	def func(grid):
		r = grid.as_('polar').r + 1e-10

		a = (L0 / r0)**(5 / 3)
		b = 2**(1 / 6) * gamma(11 / 6) / np.pi**(8 / 3)
		c = (24 / 5 * gamma(6 / 5))**(5 / 6)
		d = gamma(5 / 6) / 2**(1 / 6)
		e = (2 * np.pi * r / L0)**(5 / 6)
		f = kv(5 / 6, 2 * np.pi * r / L0)

		return Field(a * b * c * (d - e * f), grid)
	return func

def power_spectral_density_von_karman(r0, L0):
	'''Return a Field generator for the power spectral density function for Von Karman turbulence.

	Parameters
	----------
	r0 : scalar
		The Fried parameter in meters.
	L0 : scalar
		The outer scale in meters.

	Returns
	-------
	Field generator
		The power spectral density Field generator.
	'''
	def func(grid):
		u = grid.as_('polar').r + 1e-10
		u0 = 2 * np.pi / L0

		res = 0.0229 * ((u**2 + u0**2) / (2 * np.pi)**2)**(-11 / 6.) * r0**(-5 / 3)
		res[u < 1e-9] = 0

		return Field(res, grid)
	return func

def Cn_squared_from_fried_parameter(r0, wavelength=500e-9):
	'''Calculate the integrated Cn^2 for a certain Fried parameter.

	Parameters
	----------
	r0 : scalar
		The Fried parameter in meters.
	wavelength : scalar
		The wavelength in meters at which the Fried parameter is measured (default: 500nm).

	Returns
	-------
	scalar
		The integrated Cn^2 value for the atmosphere.
	'''
	k = 2 * np.pi / wavelength
	return r0**(-5. / 3) / (0.423 * k**2)

def fried_parameter_from_Cn_squared(Cn_squared, wavelength=500e-9):
	'''Calculate the Fried parameter from the integrated Cn^2.

	Parameters
	----------
	r0 : scalar
		The integrated Cn^2 value for the atmosphere.
	wavelength : scalar
		The wavelength in meters at which to calculate the Fried parameter (default: 500nm).

	Returns
	-------
	scalar
		The Fried parameter in meters.
	'''
	k = 2 * np.pi / wavelength
	return (0.423 * Cn_squared * k**2)**(-3. / 5)

def seeing_to_fried_parameter(seeing, wavelength=500e-9):
	r'''Calculate the Fried parameter from the seeing FWHM.

	This uses :math:`\theta = 0.98 \lambda / r_0`, where :math:`\theta`
	is the FWHM of the seeing, :math:`\lambda` is the wavelength and
	:math:`r_0` is the Fried parameter.

	Parameters
	----------
	seeing : scalar
		The FWHM of the seeing in arcsec.
	wavelength : scalar
		The wavelength in meters at which the seeing is measured (default: 500nm).

	Returns
	-------
	scalar
		The Fried parameter in meters at the specified wavelength.
	'''
	return 0.98 * wavelength / np.radians(seeing / 3600)

def fried_parameter_to_seeing(r0, wavelength=500e-9):
	r'''Calculate the FWHM of the seeing from the Fried parameter.

	This uses :math:`\theta = 0.98 \lambda / r_0`, where :math:`\theta`
	is the FWHM of the seeing, :math:`\lambda` is the wavelength and
	:math:`r_0` is the Fried parameter.

	Parameters
	----------
	r0 : scalar
		The Fried parameter in meters.
	wavelength : scalar
		The wavelength in meters at which the Fried parameter is measured (default: 500nm).

	Returns
	-------
	scalar
		The FWHM of the seeing in arcsec at the specified wavelength.
	'''
	return np.degrees(0.98 * wavelength / r0) * 3600



import numpy as np

from .optical_element import OpticalElement

class SimpleVibration(OpticalElement):
	def __init__(self, mode, amplitude, frequency, phase_0=0):
		self.mode = mode
		self.amplitude = amplitude
		self.frequency = frequency
		self.phase_0 = phase_0
		self.t = 0

	@property
	def frequency(self):
		return self._frequency

	@frequency.setter
	def frequency(self, frequency):
		delta_phase = 2 * np.pi * (self.frequency + self.frequency)
		self.phase_0 += delta_phase
		self._frequency = frequency

	@property
	def phase(self):
		return 2 * np.pi * self.frequency * self.t + self.phase_0

	def forward(self, wavefront):
		wf = wavefront.copy()

		wf.electric_field *= np.exp(1j * (self.mode * self.amplitude / wf.wavelength * np.sin(self.phase)))
		return wf

	def backward(self, wavefront):
		wf = wavefront.copy()

		wf.electric_field *= np.exp(-1j * (self.mode * self.amplitude / wf.wavelength * np.sin(self.phase)))
		return wf



import numpy as np
from math import sqrt
from scipy.special import eval_hermite

def index_to_hermite(i):
	'''Converts a one-dimensional mode index to the two-dimensional mode index.

	The one-dimensional mode index is converted to a two-dimensional index.
	The two-dimensional index determines the order in the x and y directions.

	Parameters
	----------
	i : int
		The one-dimensional index.

	Returns
	-------
	n : int
		The x order.
	m : int
		The y order.
	'''

	# Get the max order
	o = int((np.sqrt(8 * i + 1) - 1) / 2)

	# Get the y order
	m = int(i - o * (o + 1) / 2)

	# Get the x order
	n = o - m

	return n, m

def gaussian_hermite(n, m, mode_field_diameter=1, grid=None):
	r'''Creates a Gaussian-Hermite mode.

	This function evaluates a (n, m) order Gaussian-Hermite mode on a grid.
	The definition of the modes are the following,

	.. math::
		\exp{\left(-\frac{r^2}{w_0^2}\right)} H_n\left(\sqrt{2}\frac{x}{w_0}\right) H_m\left(\sqrt{2}\frac{y}{w_0}\right).

	Here :math:`w_0` is the mode_field_radius, which is :math:`\mathrm{MFD}/2`. This defintion follows
	the Physicists definition of the Hermite polynomials. The modes are numerical normalized to have a
	total power of 1.

	More details on the Hermite Polynomials can be found on: http://mathworld.wolfram.com/HermitePolynomial.html

	Parameters
	----------
	n : int
		The x order.
	m : int
		The y order.
	mode_field_diameter : scalar
		The mode field diameter of the Gaussian-Laguerre mode.
	grid : Grid
		The grid on which to evaluate the mode.

	Returns
	-------
	Field
		The evaluated mode.
	'''
	from ..field import Field

	if grid.is_separated and grid.is_('cartesian'):
		x = grid.x
		y = grid.y
		r2 = (x**2 + y**2) / (mode_field_diameter / 2)**2
	else:
		x, y = grid.as_('cartesian').coords
		r2 = (2 * grid.r / mode_field_diameter)**2

	# Calculate the mode.
	hg = np.exp(-r2) * eval_hermite(n, 2 * sqrt(2) * x / mode_field_diameter) * eval_hermite(m, 2 * sqrt(2) * y / mode_field_diameter)

	# Numerically normalize the mode
	hg /= np.sqrt(np.sum(np.abs(hg)**2 * grid.weights))

	return Field(hg, grid)

def gaussian_hermite_index(i, mode_field_diameter=1, grid=None):
	'''Creates a Gaussian-Hermite mode.

	This function evaluates the i-th order Gaussian-Hermite mode on a grid.

	Parameters
	----------
	i : int
		The one-dimensional mode index.
	mode_field_diameter : scalar
		The mode field diameter of the Gaussian-Hermite mode.
	grid : Grid
		The grid on which to evaluate the Gaussian-Hermite mode.

	Returns
	-------
	Field
		The evaluated Gaussian-Hermite mode.
	'''

	n, m = index_to_hermite(i)
	return gaussian_hermite(n, m, mode_field_diameter, grid)

def make_gaussian_hermite_basis(grid, num_modes, mode_field_diameter, starting_mode=0):
	'''Creates a Gaussian-Hermite mode basis.

	This function evaluates the starting_mode to num_modes + starting_modes Gaussian-Hermite modes.
	And returns a ModeBasis made out of these Gaussian-Hermite modes.

	Parameters
	----------
	grid : Grid
		The grid on which to evaluate the Gaussian-Hermite mode.
	num_modes : int
		The number of modes to create.
	mode_field_diameter : scalar
		The mode field diameter of the Gaussian-Hermite mode.
	starting_mode : int
		The starting point of the mode indices.

	Returns
	-------
	ModeBasis
		The evaluated Gaussian-Hermite modes.
	'''
	from .mode_basis import ModeBasis

	modes = [gaussian_hermite_index(i, mode_field_diameter, grid) for i in range(starting_mode, starting_mode + num_modes)]
	return ModeBasis(modes, grid)



import numpy as np

from .fast_fourier_transform import FastFourierTransform
from ..field import Field, field_dot, field_conjugate_transpose

try:
	import mkl_fft._numpy_fft as _fft_module
except ImportError:
	_fft_module = np.fft

class FourierFilter(object):
	'''A filter in the Fourier domain.

	The filtering is performed by Fast Fourier Transforms, but is quicker than
	the equivalent multiplication in the Fourier domain using the FastFourierTransform
	classes. It does this by avoiding redundant field multiplications that limit performance.

	Parameters
	----------
	input_grid : Grid
		The grid that is expected for the input field.
	transfer_function : Field generator or Field
		The transfer function to use for the filter. If this is a Field, the user is responsible
		for checking that its grid corresponds to the internal grid used by this filter. The Grid
		is not checked.
	q : scalar
		The amount of zeropadding to perform in the real domain. A value
		of 1 denotes no zeropadding. Zeropadding increases the resolution in the
		Fourier domain and therefore reduces aliasing/wrapping effects.
	'''
	def __init__(self, input_grid, transfer_function, q=1):
		fft = FastFourierTransform(input_grid, q)

		self.input_grid = input_grid
		self.internal_grid = fft.output_grid
		self.cutout = fft.cutout_input
		self.shape_in = input_grid.shape

		self.transfer_function = transfer_function

		self._transfer_function = None
		self.internal_array = None

	def _compute_functions(self, field):
		if self._transfer_function is None or self._transfer_function.dtype != field.dtype:
			if hasattr(self.transfer_function, '__call__'):
				tf = self.transfer_function(self.internal_grid)
			else:
				tf = self.transfer_function.copy()

			tf = np.fft.ifftshift(tf.shaped, axes=tuple(range(-self.input_grid.ndim, 0)))
			self._transfer_function = tf.astype(field.dtype, copy=False)

		recompute_internal_array = self.internal_array is None
		recompute_internal_array = recompute_internal_array or (self.internal_array.ndim != (field.grid.ndim + field.tensor_order))
		recompute_internal_array = recompute_internal_array or (self.internal_array.dtype != field.dtype)
		recompute_internal_array = recompute_internal_array or np.all(self.internal_array.shape[:field.tensor_order] == field.tensor_shape)

		if recompute_internal_array:
			self.internal_array = self.internal_grid.zeros(field.tensor_shape, field.dtype).shaped

	def forward(self, field):
		'''Return the forward filtering of the input field.

		Parameters
		----------
		field : Field
			The field to filter.

		Returns
		-------
		Field
			The filtered field.
		'''
		return self._operation(field, adjoint=False)

	def backward(self, field):
		'''Return the backward (adjoint) filtering of the input field.

		Parameters
		----------
		field : Field
			The field to filter.

		Returns
		-------
		Field
			The adjoint filtered field.
		'''
		return self._operation(field, adjoint=True)

	def _operation(self, field, adjoint):
		'''The internal filtering operation.

		Parameters
		----------
		field : Field
			The input field.
		adjoint : boolean
			Whether to perform a forward or adjoint filter.

		Returns
		-------
		Field
			The filtered field.
		'''
		self._compute_functions(field)

		if self.cutout is None:
			f = field.shaped
		else:
			f = self.internal_array
			f[:] = 0
			c = tuple([slice(None)] * field.tensor_order) + self.cutout
			f[c] = field.shaped

		f = _fft_module.fftn(f, axes=tuple(range(-self.input_grid.ndim, 0)))

		if (self._transfer_function.ndim - self.internal_grid.ndim) == 2:
			# The transfer function is a matrix field.
			s1 = f.shape[:-self.internal_grid.ndim] + (self.internal_grid.size,)
			f = Field(f.reshape(s1), self.internal_grid)

			s2 = self._transfer_function.shape[:-self.internal_grid.ndim] + (self.internal_grid.size,)
			tf = Field(self._transfer_function.reshape(s2), self.internal_grid)

			if adjoint:
				tf = field_conjugate_transpose(tf)

			f = field_dot(tf, f).shaped
		else:
			# The transfer function is a scalar field.
			if adjoint:
				tf = self._transfer_function.conj()
			else:
				tf = self._transfer_function

			f *= tf

		f = _fft_module.ifftn(f, axes=tuple(range(-self.input_grid.ndim, 0)))

		s = f.shape[:-self.internal_grid.ndim] + (-1,)
		if self.cutout is None:
			res = f.reshape(s)
		else:
			res = f[c].reshape(s)

		return Field(res, self.input_grid)



import numpy as np

from ..optics import Wavefront, AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward
from ..field import Field
from ..fourier import make_fourier_transform

class FraunhoferPropagator(AgnosticOpticalElement):
	'''A monochromatic perfect lens propagator.

		This implements the propagation of a wavefront through a perfect lens. The wavefront
		is assumed to be exactly in the front focal plane of the lens and is propagated to the
		back focal plane. The implementation follows [Goodman2005]_.

		.. [Goodman2005] Goodman, J.W., 2005 Introduction to Fourier optics. Roberts and Company Publishers.

		Parameters
		----------
		input_grid : Grid
			The grid on which the incoming wavefront is defined.
		output_grid : Grid
			The grid on which the outgoing wavefront is to be evaluated.
		wavelength : scalar
			The wavelength of the wavefront.
		focal_length : scalar
			The focal length of the lens system.
	'''
	def __init__(self, input_grid, output_grid, focal_length=1):
		self._input_grid = input_grid
		self._output_grid = output_grid
		self._focal_length = focal_length

		AgnosticOpticalElement.__init__(self, grid_dependent=True, wavelength_dependent=True)

	def make_instance(self, instance_data, input_grid, output_grid, wavelength):
		focal_length = self.evaluate_parameter(self.focal_length, input_grid, output_grid, wavelength)

		instance_data.uv_grid = output_grid.scaled(2 * np.pi / (focal_length * wavelength))
		instance_data.fourier_transform = make_fourier_transform(input_grid, instance_data.uv_grid)

		instance_data.norm_factor = 1 / (1j * focal_length * wavelength)

	@property
	def focal_length(self):
		return self._focal_length

	@focal_length.setter
	def focal_length(self, focal_length):
		self._focal_length = focal_length

		self.clear_cache()

	def get_input_grid(self, output_grid, wavelength):
		return self._input_grid

	def get_output_grid(self, input_grid, wavelength):
		return self._output_grid

	@make_agnostic_forward
	def forward(self, instance_data, wavefront):
		'''Propagate a wavefront forward through the lens.

		Parameters
		----------
		wavefront : Wavefront
			The incoming wavefront.

		Returns
		-------
		Wavefront
			The wavefront after the propagation.
		'''
		U_new = instance_data.fourier_transform.forward(wavefront.electric_field) * instance_data.norm_factor
		return Wavefront(Field(U_new, instance_data.output_grid), wavefront.wavelength, wavefront.input_stokes_vector)

	@make_agnostic_backward
	def backward(self, instance_data, wavefront):
		'''Propagate a wavefront backward through the lens.

		Parameters
		----------
		wavefront : Wavefront
			The incoming wavefront.

		Returns
		-------
		Wavefront
			The wavefront after the propagation.
		'''
		U_new = instance_data.fourier_transform.backward(wavefront.electric_field) / instance_data.norm_factor
		return Wavefront(Field(U_new, instance_data.input_grid), wavefront.wavelength, wavefront.input_stokes_vector)



import numpy as np
from scipy.special import jv, kn
from scipy.optimize import fsolve

from ..mode_basis import ModeBasis

def eigenvalue_equation(u, m, V):
	'''Evaluates the eigenvalue equation for a circular step-index fiber.

	Parameters
	----------
	u : scalar
		The normalized propagation constant.
	m : int
		The azimuthal order
	V : scalar
		The normalized frequency parameter of the fiber.

	Returns
	-------
	scalar
		The eigenvalue equation value
	'''
	w = np.sqrt(V**2 - u**2)
	return jv(m, u) / (u * jv(m + 1, u)) - kn(m, w) / (w * kn(m + 1, w))

def find_branch_cuts(m, V):
	'''Find all the solutions for the eigenvalue function.

	Parameters
	----------
	m : int
		The azimuthal order
	V : scalar
		The normalized frequency parameter of the fiber.

	Returns
	-------
	Tuple
		A tuple containing the solutions of the eigenvalue function. If no solutions were found returns None.
	'''
	# Make an initial rough grid
	num_steps = 501
	theta = np.linspace(np.pi * 0.499, 0, num_steps, endpoint=False)
	u = V * np.cos(theta)

	# Find the position where the eigenvalue equation goes through zero
	diff = eigenvalue_equation(u, m, V)
	fu = np.diff(np.sign(diff)) < 0
	ind = np.where(abs(fu - 1) <= 0.01)[0]

	if len(ind) > 0:
		# Refine the zero with a rootfinding algorithm
		u0 = fsolve(eigenvalue_equation, u[ind], args=(m, V))
		w0 = np.sqrt(V**2 - u0**2)
		return u0, w0
	else:
		return None


def LP_radial(m, u, w, r):
	'''Evaluates the radial profile of the LP modes.

	Parameters
	----------
	m : int
		The azimuthal order
	u : scalar
		The normalized inner propagation constant.
	w : scalar
		The normalized outer propagation constant.
	r : array_like
		The radial coordinates on which to evaluate the bessel modes.

	Returns
	-------
	array_like
		An array that contains the radial profile.
	'''
	# The scaling factor for the continuity condition
	scaling_factor = jv(m,u) /kn(m, w)

	# Find the grid inside and outside the core radius
	mask = r < 1

	# Evaluate the radial mode profile
	mode_field = np.zeros_like(r)
	mode_field[mask] = jv(m, u * r[mask])
	mode_field[~mask] = scaling_factor * kn(m, w * r[~mask])

	return mode_field

def LP_azimuthal(m, theta):
	'''Evaluates the azimuthal profile of the LP modes.

	Parameters
	----------
	m : int
		The azimuthal order
	theta : array_like
		The azimuthal coordinates on which to evaluate the cosine and sine modes.

	Returns
	-------
	array_like
		An array that contains the azimuthal profile.
	'''
	if m >= 0:
		return np.cos(m * theta)
	else:
		return np.sin(m * theta)

def make_LP_modes(grid, V_number, core_radius, return_betas=False):
	'''Make a ModeBasis out of the guided modes that are supported by a step-index fiber.

	This function solves the eigenvalue equation of for a radial step-index profile fiber.
	It calculates the mode profiles and can also return the propagation constants.

	Parameters
	----------
	grid : Grid
		The grid on which to calculate the mode basis.
	V_number : scalar
		The normalized frequency parameter of the fiber.
	core_radius : scalar
		The core radius of a step-index fiber.
	return_betas : bool
		If true this function also returns the propagation constants.
	Returns
	-------
	ModeBasis
		An ModeBasis containing all supported LP modes.
	array_like
		An array containing all propagation constants. This is only returned if return_betas is True.
	'''
	finding_new_modes = True
	m = 0
	num_modes = 0

	# scaled grid
	R, Theta = grid.scaled(1/core_radius).as_('polar').coords
	modes = []
	betas = []
	while finding_new_modes:

		solutions = find_branch_cuts(m, V_number)
		if solutions is not None:
			for ui, wi in zip(solutions[0], solutions[1]):

				radial_prodile = LP_radial(m, ui, wi, R)
				beta = np.sqrt((V_number**2 + wi**2 - ui**2) / (2 * core_radius**2))

				ms = [m,-m] if m > 0 else [m,]
				for mi in ms:
					azimutal_profile = LP_azimuthal(mi, Theta)
					mode_profile = radial_prodile * azimutal_profile

					# Normalize the mode numerically as there is no analytical normalization
					norm = np.sqrt(np.sum(mode_profile * mode_profile.conj() * grid.weights))
					mode_profile /= norm
					modes.append(mode_profile)
					betas.append(beta)
					num_modes += 1

			m += 1
		else:
			finding_new_modes = False

	# Sort the modes according to their propagation constant
	betas = np.array(betas)
	index_sorting = np.argsort(betas)[::-1]
	modes = np.array(modes)[index_sorting].T
	betas = betas[index_sorting]

	if return_betas:
		return ModeBasis(modes, grid), np.array(betas)
	else:
		return ModeBasis(modes, grid)



import numpy as np
from ..field import Field

def large_poisson(lam, thresh=1e6):
	"""
	Draw samples from a Poisson distribution, taking care of large values of `lam`.

	At large values of `lam` the distribution automatically switches to the corresponding normal distribution.
	This switch is independently decided for each expectation value in the `lam` array.

	Parameters
	----------
	lam : array_like
		Expectation value for the Poisson distribution. Must be >= 0.
	thresh : float
		The threshold at which the distribution switched from a Poisson to a normal distribution.

	Returns
	-------
	array_like
		The drawn samples from the Poisson or normal distribution, depending on the expectation value.
	"""
	large = lam > thresh
	small = ~large

	# Use normal approximation if the number of photons is large
	n = np.zeros(lam.shape)
	n[large] = np.round(lam[large] + np.random.normal(size=np.sum(large)) * np.sqrt(lam[large]))
	n[small] = np.random.poisson(lam[small], size=np.sum(small))

	if hasattr(lam, 'grid'):
		n = Field(n, lam.grid)

	return n
